package org.dd4t.thymeleaf.dialect.image.processor;

import java.util.logging.Logger;

import org.thymeleaf.IEngineConfiguration;
import org.thymeleaf.context.ITemplateContext;
import org.thymeleaf.engine.AttributeName;
import org.thymeleaf.model.IProcessableElementTag;
import org.thymeleaf.processor.element.AbstractAttributeTagProcessor;
import org.thymeleaf.processor.element.IElementTagStructureHandler;
import org.thymeleaf.standard.expression.IStandardExpression;
import org.thymeleaf.standard.expression.IStandardExpressionParser;
import org.thymeleaf.standard.expression.StandardExpressions;
import org.thymeleaf.templatemode.TemplateMode;

public class BinaryAttrProcessor extends AbstractAttributeTagProcessor {
    private static final Logger LOG = Logger.getLogger(BinaryAttrProcessor.class.getName());
    private static final String ATTR_NAME = "binary";
    private static final int PRECEDENCE = 10000;

    private SizedImageFactory sizedImageFactory;

    /**
     * Image attribute processor constructor.
     * 
     * @param propertiesService
     */
    public BinaryAttrProcessor(final String dialectPrefix, SizedImageFactory sizedImageFactory) {
        super(TemplateMode.HTML, // This processor will apply only to HTML mode
                dialectPrefix, // Prefix to be applied to name for matching
                null, // No tag name: match any tag name
                false, // No prefix to be applied to tag name
                ATTR_NAME, // Name of the attribute that will be matched
                true, // Apply dialect prefix to attribute name
                PRECEDENCE, // Precedence (inside dialect's own precedence)
                true); // Remove the matched attribute afterwards
        this.sizedImageFactory = sizedImageFactory;
    }

    @Override
    protected void doProcess(ITemplateContext context, IProcessableElementTag tag, AttributeName attributeName, String attributeValue, IElementTagStructureHandler structureHandler) {
        final IEngineConfiguration configuration = context.getConfiguration();
        final IStandardExpressionParser parser = StandardExpressions.getExpressionParser(configuration);
        final IStandardExpression expressionImageUrl = parser.parseExpression(context, attributeValue);
        String alt = "";
        String title = "";

        if (tag.hasAttribute("alt")) {
            IStandardExpression expressionAlt = parser.parseExpression(context, tag.getAttributeValue("alt"));
            alt = (String) expressionAlt.execute(context);
        }
        if (tag.hasAttribute("title")) {
            IStandardExpression expressionTitle = parser.parseExpression(context, tag.getAttributeValue("title"));
            title = (String) expressionTitle.execute(context);
        }
        TridionBinaryViewModel binary = (TridionBinaryViewModel) expressionImageUrl.execute(context);
        if (binary == null) {
            return;
        }

        String classAttr = tag.getAttributeValue("class");

        String widthAttr = tag.getAttributeValue("width");
        String heightAttr = tag.getAttributeValue("height");
        int width;
        int height;
        if (widthAttr == null) {
            width = 0;
        } else {
            width = Integer.parseInt(widthAttr);
        }
        if (heightAttr == null) {
            height = 0;
        } else {
            height = Integer.parseInt(heightAttr);
        }

        StringBuffer result = new StringBuffer();
        result.append("<img ");
        if (alt != null && !alt.isEmpty()) {
            result.append("alt=\"" + alt + "\" ");
        }
        if (title != null && !title.isEmpty()) {
            result.append("title=\"" + title + "\" ");
        }
        if (classAttr != null && !classAttr.isEmpty()) {
            result.append("class=\"" + classAttr + "\" ");
        }
        if (width > 0 || height > 0) {
            SizedImage sizedImage = sizedImageFactory.getSizedImage(binary, width, height);
            sizedImage = sizedImage.resize();

            if (sizedImage != null) {
                result.append("src=\"" + sizedImage.getSizedImageUrl() + "\"");
            }
        } else {
            SizedImage sizedImage = sizedImageFactory.getSizedImage(binary, width, height);
            if (sizedImage != null) {
                result.append("src=\"" + sizedImage.getUrl() + "\"");
            }
        }
        result.append(" />");

        structureHandler.setBody(result, false);
    }
}