package org.dd4t.thymeleaf.dialect.image.processor;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.AreaAveragingScaleFilter;
import java.awt.image.BufferedImage;
import java.awt.image.FilteredImageSource;
import java.awt.image.ImageFilter;
import java.awt.image.ImageProducer;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.Iterator;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.plugins.jpeg.JPEGImageWriteParam;
import javax.imageio.stream.ImageOutputStream;
import javax.swing.ImageIcon;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.SerializationException;
import org.dd4t.contentmodel.Binary;
import org.dd4t.contentmodel.impl.BinaryDataImpl;
import org.dd4t.contentmodel.impl.BinaryImpl;
import org.dd4t.core.exceptions.ItemNotFoundException;
import org.dd4t.core.resolvers.PublicationResolver;
import org.dd4t.core.services.PropertiesService;
import org.dd4t.core.util.TCMURI;

import com.sdl.web.api.content.BinaryContentRetriever;
import com.sdl.web.api.dynamic.BinaryContentRetrieverImpl;
import com.sdl.web.api.meta.WebBinaryMetaFactory;
import com.sdl.web.api.meta.WebBinaryMetaFactoryImpl;
import com.tridion.data.BinaryData;
import com.tridion.meta.BinaryMeta;

public class SizedImage {

    private static final Logger LOG = Logger.getLogger(BinaryAttrProcessor.class.getName());
    private static final BinaryContentRetriever BINARY_CONTENT_RETRIEVER = new BinaryContentRetrieverImpl();
    private static final WebBinaryMetaFactory WEB_BINARY_META_FACTORY = new WebBinaryMetaFactoryImpl();
    private static final int LANDSCAPE_DIVIDOR = 5;
    private static final int PORTRAIT_DIVIDOR = 5;
    private String diskImagePath = "site/images";
    private String imagePath = "images";
    private TridionBinaryViewModel image;
    private String url;
    private int focalPoint;
    private int width;
    private int height;
    private int originalWidth;
    private int originalHeight;
    private String sizedImageUrl;
    private String extension;
    private String fileName;
    private String hash;
    private String dirs;
    private String originalImageLocation;
    private String decodedFileName;
    private float jpgQuality = 0.9F;
    private PublicationResolver publicationResolver;

    public SizedImage(String url, int width, int height, PropertiesService propertiesService, PublicationResolver publicationResolver) {
        this.publicationResolver = publicationResolver;

        this.diskImagePath = propertiesService.getProperty("images.physicalpath");
        this.imagePath = propertiesService.getProperty("images.url");

        this.url = url;
        this.width = width;
        this.height = height;

        loadImageUrl();
    }

    public SizedImage(TridionBinaryViewModel binary, int width, int height, PropertiesService propertiesService, PublicationResolver publicationResolver) {
        this.publicationResolver = publicationResolver;

        this.diskImagePath = propertiesService.getProperty("images.physicalpath");
        this.imagePath = propertiesService.getProperty("images.url");

        this.image = binary;
        this.width = width;
        this.height = height;

        loadImage();
    }

    private void loadImage() {
        this.extension = image.getFileExtension();

        String[] urlPath = image.getUrl().split("/");
        this.fileName = urlPath[urlPath.length - 1].substring(0, (urlPath[urlPath.length - 1].length() - (extension.length() + 1)));

        this.hash = MD5.hash(focalPoint + "_" + image.getSize() + "_" + image.getLastModified() + "_" + image.getLastPublishDate() + "_" + fileName);

        this.dirs = imageDirs(hash.substring(0, 3));

        try {
            decodedFileName = URLDecoder.decode(fileName, StandardCharsets.UTF_8.name());
        } catch (UnsupportedEncodingException e1) {
            LOG.log(Level.WARNING, "unexpected error: unsupported encoding");
            decodedFileName = fileName;
        }

        this.originalImageLocation = dirs + "/" + hash + "_" + decodedFileName + "." + extension;

        File originalImageLocationFile = new File(originalImageLocation);

        if (!originalImageLocationFile.exists()) {
            storeOriginalOnDisk(image.getUrl(), originalImageLocationFile, publicationResolver);

            try (FileInputStream fis = new FileInputStream(originalImageLocationFile)) {
                BufferedImage image;
                image = ImageIO.read(fis);
                setOriginalWidth(image.getWidth());
                setOriginalHeight(image.getHeight());
            } catch (FileNotFoundException e) {
                LOG.log(Level.SEVERE, e.getMessage(), e);
            } catch (IOException e) {
                LOG.log(Level.SEVERE, e.getMessage(), e);
            } catch (Exception e) {
                LOG.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        setUrl(encodeFileName("/" + originalImageLocation.replace(diskImagePath, imagePath)));
    }

    private void loadImageUrl() {
        this.extension = url.substring(url.lastIndexOf(".") + 1);

        String[] urlPath = url.split("/");
        this.fileName = urlPath[urlPath.length - 1].substring(0, (urlPath[urlPath.length - 1].length() - (extension.length() + 1)));

        this.hash = MD5.hash(focalPoint + width + height + "_" + fileName);

        this.dirs = imageDirs(hash.substring(0, 3));

        try {
            decodedFileName = URLDecoder.decode(fileName, StandardCharsets.UTF_8.name());
        } catch (UnsupportedEncodingException e1) {
            LOG.log(Level.WARNING, "unexpected error: unsupported encoding");
            decodedFileName = fileName;
        }

        this.originalImageLocation = dirs + "/" + hash + "_" + decodedFileName + "." + extension;

        File originalImageLocationFile = new File(originalImageLocation);

        if (!originalImageLocationFile.exists()) {
            originalImageLocationFile = storeOriginalOnDisk(url, originalImageLocationFile, publicationResolver);

            // if binary is not accessible
            if (originalImageLocationFile == null) {
                setSizedImageUrl("");
                setUrl("");
                originalImageLocation = "";
                return;
            }

            try (FileInputStream fis = new FileInputStream(originalImageLocationFile)) {
                BufferedImage image;
                image = ImageIO.read(fis);
                setOriginalWidth(image.getWidth());
                setOriginalHeight(image.getHeight());
            } catch (FileNotFoundException e) {
                LOG.log(Level.SEVERE, e.getMessage(), e);
            } catch (IOException e) {
                LOG.log(Level.SEVERE, e.getMessage(), e);
            } catch (Exception e) {
                LOG.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        setUrl(encodeFileName("/" + originalImageLocation.replace(diskImagePath, imagePath)));
    }

    public SizedImage resize() {
        if (originalImageLocation == null) {
            return null;
        }

        File originalImageLocationFile = new File(originalImageLocation);
        if (!originalImageLocationFile.exists()) {
            LOG.log(Level.WARNING, "cannot resize image, original file '" + originalImageLocation + "' not found; using original image instead");
            setSizedImageUrl(this.image.getUrl());
            return this;
        }

        String sizedImageLocation = dirs + "/" + hash + "_" + fileName + "_" + getWidth() + "x" + getHeight() + "." + extension;
        // LOG.log(Level.INFO, "sizedImageLocation=" + sizedImageLocation);

        File sizedImageLocationFile = new File(sizedImageLocation);
        if (!sizedImageLocationFile.exists()) {
            // LOG.log(Level.INFO, "generating sizedImage=" + sizedImageLocationFile.getPath());

            try {
                if (extension.toLowerCase().equals("jpg") || extension.toLowerCase().equals("jpeg")) {
                    // LOG.log(Level.INFO, "generating JPG");
                    resizeAndCropToJPG(new FileInputStream(originalImageLocationFile), new FileOutputStream(sizedImageLocationFile), width, height, getJPGQuality(), focalPoint);
                } else if (extension.toLowerCase().equals("png")) {
                    // LOG.log(Level.INFO, "generating PNG");
                    resizeAndCropToPNG(new FileInputStream(originalImageLocationFile), new FileOutputStream(sizedImageLocationFile), width, height, focalPoint);
                } else {
                    // LOG.log(Level.SEVERE, "Error resizing image " + originalImageLocation + ": unknown extension " + extension);
                }
            } catch (FileNotFoundException e) {
                LOG.log(Level.SEVERE, e.getMessage());
            } catch (IOException e) {
                sizedImageLocationFile.delete();
                LOG.log(Level.SEVERE, e.getMessage());
            }
        } else {
            // LOG.log(Level.INFO, "sizedImageLocation is already cached=" + sizedImageLocationFile.getPath());
        }

        setHeight(height);
        setWidth(width);
        setSizedImageUrl(encodeFileName("/" + sizedImageLocation.replace(diskImagePath, imagePath)));
        return this;
    }

    public static String encodeFileName(String url) {
        String fileName = FilenameUtils.getName(url);
        try {
            return url.replace(fileName, URLEncoder.encode(fileName, StandardCharsets.UTF_8.name()));
        } catch (UnsupportedEncodingException e) {
            LOG.warning("unexpected error: unsupported encoding");
            return url;
        }
    }

    private static File storeOriginalOnDisk(String url, File file, PublicationResolver publicationResolver) {
        try {
            Binary binary = getBinaryByURL(url, publicationResolver.getPublicationId());
            FileOutputStream fileOutputStream = new FileOutputStream(file);
            InputStream is = binary.getBinaryData().getInputStream();
            byte[] buffer = new byte[1024];
            int len1 = 0;
            while ((len1 = is.read(buffer)) > 0) {
                fileOutputStream.write(buffer, 0, len1);
            }
            fileOutputStream.close();
            is.close();
        } catch (SerializationException e) {
            file = null;
            LOG.log(Level.SEVERE, e.getMessage(), e);
        } catch (ItemNotFoundException e) {
            file = null;
            LOG.log(Level.SEVERE, e.getMessage(), e);
        } catch (FileNotFoundException e) {
            file = null;
            LOG.log(Level.SEVERE, e.getMessage(), e);
        } catch (IOException e) {
            file = null;
            LOG.log(Level.SEVERE, e.getMessage(), e);
        }
        return file;
    }

    public static void resizeAndCropToJPG(InputStream in, OutputStream out, int width, int height, float jpgQuality, int focalPoint) throws IOException {
        BufferedImage image = ImageIO.read(in);
        BufferedImage finalImage = resizeAndCrop(width, height, image, false, focalPoint);
        if (jpgQuality < 0 || jpgQuality > 1) {
            // default to 1 if an illegal value is used
            LOG.log(Level.WARNING, "illegal jpg quality level of " + jpgQuality + ". The value should be between 0 and 1. Using " + "default of 1.");
            jpgQuality = 1;
        }
        try {
            writeAsCompressedJpeg(finalImage, out, jpgQuality);
        } finally {
            out.close();
            in.close();
        }
    }

    public static void resizeAndCropToPNG(InputStream in, OutputStream out, int width, int height, int focalPoint) throws IOException {
        BufferedImage image = ImageIO.read(in);
        BufferedImage finalImage = resizeAndCrop(width, height, image, true, focalPoint);
        try {
            ImageIO.write(finalImage, "png", out);
        } finally {
            out.close();
            in.close();
        }
    }

    public static Binary getBinaryByURL(final String url, final int publication) throws ItemNotFoundException, SerializationException {
        final BinaryMeta binaryMeta = getBinaryMetaByURL(url, publication);

        if (binaryMeta == null) {
            throw new ItemNotFoundException("Unable to find binary content by URL '" + url + "' and publication '" + publication + "'.");
        }
        TCMURI binaryUri = new TCMURI(binaryMeta.getPublicationId(), (int) binaryMeta.getId(), 16);
        return getBinary(binaryUri, binaryMeta);
    }

    private static BinaryMeta getBinaryMetaByURL(String url, int publication) throws ItemNotFoundException {
        final BinaryMeta binaryMeta = WEB_BINARY_META_FACTORY.getMetaByURL(publication, url);
        if (binaryMeta == null) {
            throw new ItemNotFoundException("Unable to find binary by url '" + url + "' and publication '" + publication + "'.");
        }
        return binaryMeta;
    }

    private static Binary getBinary(final TCMURI binaryUri, final BinaryMeta binaryMeta) throws ItemNotFoundException {
        if (binaryMeta != null) {
            final BinaryImpl binary = new BinaryImpl();

            binary.setId(binaryUri.toString());
            binary.setUrlPath(binaryMeta.getURLPath());
            binary.setMimeType(binaryMeta.getType());

            final BinaryDataImpl binaryDataBytes = new BinaryDataImpl();
            binaryDataBytes.setBytes(getBinaryContentById(binaryUri.getItemId(), binaryUri.getPublicationId()));
            binary.setBinaryData(binaryDataBytes);
            return binary;
        }
        return null;
    }

    public static byte[] getBinaryContentById(int id, int publication) throws ItemNotFoundException {

        final BinaryData binaryData = BINARY_CONTENT_RETRIEVER.getBinary(publication, id);

        if (binaryData == null || binaryData.getDataSize() == 0) {
            throw new ItemNotFoundException("Unable to find binary content by id: tcm:" + publication + "-" + id);
        }

        try {
            return binaryData.getBytes().clone();
        } catch (IOException e) {
            // TODO: wrap in provider exception
            throw new ItemNotFoundException(e);
        }
    }

    public static void writeAsCompressedJpeg(BufferedImage image, OutputStream out, float compressionQuality) throws IOException {
        // Find a jpeg writer
        ImageWriter writer = null;
        Iterator<ImageWriter> iter = ImageIO.getImageWritersByFormatName("jpg");
        if (iter.hasNext()) {
            writer = iter.next();
        }
        ImageOutputStream ios = ImageIO.createImageOutputStream(out);
        try {
            writer.setOutput(ios);

            // Set the compression quality
            ImageWriteParam iwparam = new JPEGImageWriteParam(Locale.getDefault());
            iwparam.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
            iwparam.setCompressionQuality(compressionQuality);

            // Write the image
            writer.write(null, new IIOImage(image, null, null), iwparam);
        } finally {
            // Cleanup
            ios.flush();
            writer.dispose();
            ios.close();
        }
    }

    public void setUrl(String url) {
        this.url = url;
    }

    public void setWidth(int width) {
        this.width = width;
    }

    public int getWidth() {
        return this.width;
    }

    public void setHeight(int height) {
        this.height = height;
    }

    public int getHeight() {
        return this.height;
    }

    public void setOriginalWidth(int originalWidth) {
        this.originalWidth = originalWidth;
    }

    public int getOriginalWidth() {
        return this.originalWidth;
    }

    public void setOriginalHeight(int originalHeight) {
        this.originalHeight = originalHeight;
    }

    public int getOriginalHeight() {
        return this.originalHeight;
    }

    private void setSizedImageUrl(String sizedImageUrl) {
        this.sizedImageUrl = sizedImageUrl;
    }

    public float getJPGQuality() {
        return this.jpgQuality;
    }

    private void setJPGQuality(float jpgQuality) {
        this.jpgQuality = jpgQuality;
    }

    public String getSizedImageUrl() {
        return this.sizedImageUrl;
    }

    private String imageDirs(String dirs) {
        dirs = diskImagePath + "/" + new StringBuffer(dirs).insert(1, "/").insert(3, "/").toString();
        File fileDirs = new File("" + dirs);
        if (!fileDirs.isDirectory()) {
            boolean success = (fileDirs.mkdirs());
            if (!success) {
                LOG.log(Level.SEVERE, "cannot create dirs=" + dirs);
                // TODO: trigger a 500 exception
            }
        }
        return dirs;
    }

    private static BufferedImage resizeAndCrop(int targetX, int targetY, BufferedImage image, boolean preserveAlpha, int focalPoint) {
        // work around for http: // bugs.sun.com/bugdatabase/view_bug.do?bug_id=4886071
        // this makes the code about 20-30x faster.
        if (!image.getColorModel().getColorSpace().isCS_sRGB()) {
            BufferedImage image2 = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_RGB);
            Graphics g = image2.getGraphics();
            g.drawImage(image, 0, 0, null);
            g.dispose();
            image = image2;
        }
        int origX = image.getWidth();
        int origY = image.getHeight();

        // don't blow image up!
        if (targetY > origY || targetX > origX) {
            targetY = origY;
            LOG.log(Level.INFO, "New height(" + targetY + ") bigger then original height(" + origY + ") switching back to original height.");
            targetX = origX;
            LOG.log(Level.INFO, "New width(" + targetX + ") bigger then original width(" + origX + ") switching back to original width.");
        }

        // determine the original and target aspectratio to decide how to crop
        float originalAspectRatio = ((float) origY) / origX;
        float targetAspectRatio = ((float) targetY) / targetX;

        if (targetAspectRatio == 0.0 || Float.isInfinite(targetAspectRatio)) {
            if (origY < origX) {
                if (targetY == 0) {
                    targetAspectRatio = originalAspectRatio - 0.00001F;
                } else {
                    targetAspectRatio = originalAspectRatio + 0.00001F;
                }
            } else if (origY > origX) {
                if (targetX == 0) {
                    targetAspectRatio = originalAspectRatio - 0.00001F;
                } else {
                    targetAspectRatio = originalAspectRatio + 0.00001F;
                }
            } else {
                targetAspectRatio = originalAspectRatio;
            }

            if (targetX == 0) {
                targetX = Math.round(targetY * targetAspectRatio);
            } else {
                targetY = Math.round(targetX * targetAspectRatio);
            }
        }

        BufferedImage finalImage;
        if (originalAspectRatio < targetAspectRatio) { // crop from the side
                                                       // (original is wider)
            int newY = targetY;
            float scaleFactor = ((float) newY) / origY;
            int newX = Math.round(origX * scaleFactor); // size width
                                                        // proportionally

            int focalMultiplier = ((focalPoint - 1) % LANDSCAPE_DIVIDOR);

            BufferedImage scaledImage = resizeImage(image, newX, newY, preserveAlpha);

            int x = (((newX - targetX) / (LANDSCAPE_DIVIDOR - 1)) * focalMultiplier);
            if (x <= 0) {
                x = 0;
            }
            finalImage = scaledImage.getSubimage(x, 0, targetX, targetY);
        } else if (originalAspectRatio > targetAspectRatio) { // crop from top
                                                              // (original is
                                                              // higher)
            int newX = targetX;
            float scaleFactor = ((float) newX) / origX;
            int newY = Math.round(origY * scaleFactor); // size width
                                                        // proportionally

            int focalMultiplier = (((focalPoint - 1) / LANDSCAPE_DIVIDOR) % PORTRAIT_DIVIDOR);

            BufferedImage scaledImage = resizeImage(image, newX, newY, preserveAlpha);

            int y = (((newY - targetY) / (PORTRAIT_DIVIDOR - 1)) * focalMultiplier);
            if (y <= 0) {
                y = 0;
            }
            finalImage = scaledImage.getSubimage(0, y, targetX, targetY);
        } else { // aspect ratios are the same, don't crop
            finalImage = resizeImage(image, targetX, targetY, preserveAlpha);
        }
        return finalImage;
    }

    private static BufferedImage resizeImage(BufferedImage sourceImage, int width, int height, boolean preserveAlpha) {
        ImageFilter filter = new AreaAveragingScaleFilter(width, height);
        ImageProducer producer = new FilteredImageSource(sourceImage.getSource(), filter);
        Image resizedImage = Toolkit.getDefaultToolkit().createImage(producer);
        return toBufferedImage(resizedImage, preserveAlpha);
    }

    private static BufferedImage toBufferedImage(Image i, boolean preserveAlpha) {
        int type = preserveAlpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;
        Image image = new ImageIcon(i).getImage();
        BufferedImage bufferedImage = new BufferedImage(image.getWidth(null), image.getHeight(null), type);
        Graphics2D g = bufferedImage.createGraphics();
        if (preserveAlpha) {
            g.setComposite(AlphaComposite.Src);
        }
        g.setBackground(Color.white);
        g.fillRect(0, 0, image.getWidth(null), image.getHeight(null));
        g.drawImage(image, 0, 0, null);
        g.dispose();
        return bufferedImage;
    }

    public String getUrl() {
        return url;
    }
}